/******************************************************
*
* copyright by H.G.Kleta, 2002, 2006
*
* if you wish to use these sources, please contact
*
* fhw at indigo2 dot de
*
******************************************************/

#include "mt.h"


/******************************************************
*
* variables
*
******************************************************/

/* pointer to actual fieldfuntion */

Coord (*f)(Coord x, Coord y, Coord z);


/******************************************************
*
* Coord f(Coord x, Coord y, Coord z)
*
* function returns distance of point p to isosurface
*
* various implementations...
*
******************************************************/


/* sphere */

Coord f1(Coord x, Coord y, Coord z)
{
  Coord r = 5.0;
  return (sqrt(x*x + y*y + z*z) - r);
}/*f1()*/


/* ellipsoid */

Coord f2(Coord x, Coord y, Coord z)
{
  Coord r = 3.0;
  return (sqrt(x*x/r + y*y + z*z*r) - r);
}/*f2()*/


/* super-ellipsoid */

Coord f3(Coord x, Coord y, Coord z)
{
  Coord r = 4.0;
  return (sqrt(abs(x*x*x)/r + abs(y*y*y)/r + abs(z*z*z)/r) - r);
}/*f3()*/


/* hot-dog */

Coord f4(Coord x, Coord y, Coord z)
{
  Coord r      = 2.0;
  Coord length = 3.0;

  if((x <= length) && (x >= -length))
    {
/* middle */
      return (sqrt(z*z + y*y) - r);
    }/*if*/
  else
    {
      if(x < -length)
/* left */
        return (sqrt((x+length)*(x+length) + y*y + z*z) - r);
      else
/* right */
        return (sqrt((x-length)*(x-length) + y*y + z*z) - r);
    }/*else*/
}/*f4()*/


/* buoy */

Coord f5(Coord x, Coord y, Coord z)
{
  Coord length = 5.0;
  Coord r      = 0.8 * length;

  if((x <= (0.1 * length)) && (x >= -(0.1 * length)))
    {
/* middle */
      return (sqrt(z*z + y*y) - r);
    }/*if*/
  else
    {
      if(x < -(0.1 * length))
        {
/* left */
          return (sqrt(y*y + z*z) - x - (0.9 * length));
        }/*if*/
      else
        {
/* right */
          return (sqrt(y*y + z*z) + x - (0.9 * length));
        }/*else*/
    }/*else*/
}/*f5()*/


/* atom */

Coord f6(Coord x, Coord y, Coord z)
{
  Coord r      = 1.5;
  Coord length = 5 * r;

/* x: left / right */
/* left */
  if(x < -length)
    {
      return (sqrt((x+length)*(x+length) + y*y + z*z) - r);
    }/*if*/
/* right */
  if(x > length)
    {
      return (sqrt((x-length)*(x-length) + y*y + z*z) - r);
    }/*if*/
/* arms in between */
  if( ((x <= length) && (x > (1.75 * r))) ||
      ((x >= -length) && (x < -(1.75 * r))) )
    {
      return (sqrt(y*y + z*z) - r);
    }/*if*/
  else
    {
/* y: bottom / top */
/* bottom */
      if(y < -length)
        {
          return (sqrt(x*x + (y+length)*(y+length) + z*z) - r);
        }/*if*/
/* top */
      if(y > length)
        {
          return (sqrt(x*x + (y-length)*(y-length) + z*z) - r);
        }/*if*/
/* arms in between */
      if( ((y <= length) && (y > (1.75 * r))) ||
          ((y >= -length) && (y < -(1.75 * r))) )
        {
          return (sqrt(x*x + z*z) - r);
        }/*if*/
      else
        {
/* z: back / front */
/* back */
          if(z < -length)
            {
              return (sqrt(x*x + y*y + (z+length)*(z+length)) - r);
            }/*if*/
/* front */
          if(z > length)
            {
              return (sqrt(x*x + y*y + (z-length)*(z-length)) - r);
            }/*if*/
/* arms in between */
          if( ((z <= length) && (z > (1.75 * r))) ||
              ((z >= -length) && (z < -(1.75 * r))) )
            {
              return (sqrt(x*x + y*y) - r);
            }/*if*/
/* center */
          else
            {
              return (sqrt(x*x + y*y + z*z) - (2 * r));
            }/*else: (not z)*/
        }/*else: (not y)*/
    }/*else: (not x)*/
}/*f6()*/


/* torus */

Coord f7(Coord x, Coord y, Coord z)
{
/* normal of torus plane */
  Vector3D normal = {0.0, 0.0, 1.0};
/* origin of torus */
  Point3D m   = {0.0, 0.0, 0.0};
/* radii of torus */
  Coord R     = 5.0;
  Coord r     = 2.0;
/* middle of torus-slice */
  Point3D c;
/* projection of point onto torus-plane */
  Point3D p1;
/* same as vector */
  Vector3D vp1;
/* "height" of p above torus plane */
  Coord height;
/* return value */
  Coord result;

/* projection of p onto plane of torus */
  height = (x - m[X])*normal[X] + (y - m[Y])*normal[Y] + (z - m[Z])*normal[Z];
  p1[X] = x - height * normal[X];
  p1[Y] = y - height * normal[Y];
  p1[Z] = z - height * normal[Z];

/* normalized vector m -> p1 */
  vp1[X] = p1[X] - m[X];
  vp1[Y] = p1[Y] - m[Y];
  vp1[Z] = p1[Z] - m[Z];
  NormalizeVector(vp1,vp1);

/* center of torus */
  c[X] = vp1[X] * R;
  c[Y] = vp1[Y] * R;
  c[Z] = vp1[Z] * R;

/* from now on: like a sphere, since a torus is a rotated circle... */
  result = sqrt((x-c[X])*(x-c[X]) + (y-c[Y])*(y-c[Y]) + (z-c[Z])*(z-c[Z])) - r;

  return result;
}/*f7()*/


/* egg */

Coord f8(Coord x, Coord y, Coord z)
{
/* radii */
  Coord r_g = 4.0;
  Coord r_h = 4.0;

/* center */
  Point3D G = { 0.0, 3.0, 0.0};
  Point3D H = { 0.0,-3.0, 0.0};

/* results of fieldfunctions */
  Coord result_g;
  Coord result_h;

/* calculate results of fieldfunctions */
  result_g = sqrt((x-G[X])*(x-G[X]) + 
                  (y-G[Y])*(y-G[Y]) + 
                  (z-G[Z])*(z-G[Z])) - r_g;
  result_h = sqrt((x-H[X])*(x-H[X]) + 
                  (y-H[Y])*(y-H[Y]) + 
                  (z-H[Z])*(z-H[Z])) - r_h;

/* define influence of fieldfunctions */
  result_g = 0.4 * result_g;
  result_h = 0.6 * result_h;

  return (result_g + result_h);
}/*f8()*/


/* peanut */

Coord f9(Coord x, Coord y, Coord z)
{
/* radii */
  Coord r_g = 4.0;
  Coord r_h = 4.0;

/* center */
  Point3D G = { 0.0,-3.5, 0.0};
  Point3D H = { 0.0, 3.5, 0.0};

/* results of fieldfunctions */
  Coord result_g;
  Coord result_h;

/* p as Point3D - needed for usage of GetPointDistance() */
  Point3D P;
  P[X] = x;
  P[Y] = y;
  P[Z] = z;

/* calculate results of fieldfunctions */
  result_g = sqrt((x-G[X])*(x-G[X]) + 
                  (y-G[Y])*(y-G[Y]) + 
                  (z-G[Z])*(z-G[Z])) - r_g;
  result_h = sqrt((x-H[X])*(x-H[X]) + 
                  (y-H[Y])*(y-H[Y]) + 
                  (z-H[Z])*(z-H[Z])) - r_h;

/* influence of fieldfunctions is dependent on distance point-seed */
  result_g = result_g / GetPointDistance(G,P);
  result_h = result_h / GetPointDistance(H,P);

  return (result_g + result_h);
}/*f9()*/


/******************************************************
*
* void NormalizeVector(Vector3D v, Vector3D result)
*
* function calculates normalized version of v
*
*          return: result = vector with lenght = 1
*                           and direction from v
*
******************************************************/

void NormalizeVector(Vector3D v, Vector3D result)
{
  Coord absoluteVectorValue;

/* calculate length of vector */
  absoluteVectorValue = sqrt(v[X] * v[X] + 
                             v[Y] * v[Y] + 
                             v[Z] * v[Z]);

/* safety first: prevent division by zero by returning standard vector */
  if(absoluteVectorValue==0)
    {
      result[X] = 0.57735;
      result[Y] = 0.57735;
      result[Z] = 0.57735;
    }/*if*/
/* return normalized vector */
  else
    {
      result[X] = result[X] / absoluteVectorValue;
      result[Y] = result[Y] / absoluteVectorValue;
      result[Z] = result[Z] / absoluteVectorValue;
    }/*else*/
}/*NormalizeVector()*/


/******************************************************
*
* void GetVectorProduct(Vector3D v1, Vector3D v2, Vector3D result)
*
* function caltulates crossproduct of v1 and v2
*
*          return: result = v1 x v2
*
*                    |  x   y   z  |
*          v1 x v2 = | v1x v1y v1z |
*                    | v2x v2y v2z |
*
******************************************************/

void GetVectorProduct(Vector3D v1, Vector3D v2, Vector3D result)
{
  result[X] = (v1[Y] * v2[Z]) - (v2[Y] * v1[Z]);
  result[Y] = (v1[Z] * v2[X]) - (v2[Z] * v1[X]);
  result[Z] = (v1[X] * v2[Y]) - (v2[X] * v1[Y]);
}/*GetVectorProduct()*/


/******************************************************
*
* Coord GetScalarProduct(Vector3D v1, Vector3D v2)
*
* function calculates scalarproduct of v1 and v2
*
*          return: GetScalarProduct(v1,v2) = v1 . v2
*
*          v1 . v2 = v1x * v2x + v1y * v2y + v1z * v2z
*
******************************************************/

Coord GetScalarProduct(Vector3D v1, Vector3D v2)
{
  Coord result;

  result = (v1[X] * v2[X]) + (v1[Y] * v2[Y]) + (v1[Z] * v2[Z]);

  return result;
}/*GetScalarProduct()*/


/******************************************************
*
* Coord GetEdgeAngle(EdgeListElement *e1, EdgeListElement *e2)
*
* function calculates angle between e1 and e2
*
*          return: -1 : Angle > 180
*                  else cos(angle between e1 and e2)
*
******************************************************/

Coord GetEdgeAngle(EdgeListElement *e1, EdgeListElement *e2)
{
  Coord normalAngle;
  Coord result;

  Vector3D v1,v2,normal,n1;

  v1[X] = e1->edge.i[X] - e1->edge.j[X];
  v1[Y] = e1->edge.i[Y] - e1->edge.j[Y];
  v1[Z] = e1->edge.i[Z] - e1->edge.j[Z];

  v2[X] = e2->edge.j[X] - e2->edge.i[X];
  v2[Y] = e2->edge.j[Y] - e2->edge.i[Y];
  v2[Z] = e2->edge.j[Z] - e2->edge.i[Z];

  GetNormal(e1->edge.i,e1->edge.j,e1->edge.k,normal);
  GetVectorProduct(normal,v1,n1);

  normalAngle = GetScalarProduct(v2,n1) / 
                (sqrt(v2[X]*v2[X] + v2[Y]*v2[Y] + v2[Z]*v2[Z]) *
                 sqrt(n1[X]*n1[X] + n1[Y]*n1[Y] + n1[Z]*n1[Z]) );

  if(normalAngle >= 0)
    {
      result = GetScalarProduct(v1,v2) / 
               (sqrt(v1[X]*v1[X] + v1[Y]*v1[Y] + v1[Z]*v1[Z]) *
                sqrt(v2[X]*v2[X] + v2[Y]*v2[Y] + v2[Z]*v2[Z]) );
    }/*if*/
  else
    {
      result = -1;
    }/*else*/

  return result;
}/*GetEdgeAngle()*/


/******************************************************
*
* void GetGradient(Point3D p, Vector3D result)
*
* function calculates gradient of point p in scalar field
*
*          return: result = gradient of p in scalar field
*                           can be used as normal
*
*          result x = f(x + dx, y, z) - f(x - dx, y, z)
*          result y = f(x, y + dy, z) - f(x, y - dy, z)
*          result z = f(x, y, z + dz) - f(x, y, z - dz)
*
*          result = result / abs(result)
*
******************************************************/

void GetGradient(Point3D p, Vector3D result)
{
  Coord delta = 0.01;

  result[X] = f(p[X] + delta, p[Y], p[Z]) -
              f(p[X] - delta, p[Y], p[Z]);
  result[Y] = f(p[Y], p[Y] + delta, p[Z]) -
              f(p[Y], p[Y] - delta, p[Z]);
  result[Z] = f(p[Z], p[Y], p[Z] + delta) -
              f(p[Z], p[Y], p[Z] - delta);

  NormalizeVector(result, result);
}/*GetGradient()*/


/******************************************************
*
* void GetNormal(Point3D p1, Point3D p2, Point3D p3 , Vector3D result)
*
* function calculates normal of triangle described by p1,p2,p3
*
*          return: result = normal of triangle described by p1,p2,p3
*
******************************************************/

void GetNormal(Point3D p1, Point3D p2, Point3D p3, Vector3D result)
{
  result[X] = (p2[Y]-p1[Y])*(p3[Z]-p2[Z])-(p3[Y]-p2[Y])*(p2[Z]-p1[Z]);
  result[Y] = (p2[Z]-p1[Z])*(p3[X]-p2[X])-(p3[Z]-p2[Z])*(p2[X]-p1[X]);
  result[Z] = (p2[X]-p1[X])*(p3[Y]-p2[Y])-(p3[X]-p2[X])*(p2[Y]-p1[Y]);
}/*GetNormal()*/


/******************************************************
*
* void GetNewPoint(Point3D i, Point3D j, Point3D k, Coord l, Point3D x)
*
* function calculates point x in plane i,j,k (counterclockwise ->
*          mathmatical positive viewed from top) perpendicular to edge i,j
*          with distance l from e (middle between i and j)
*
*          return: x
*
*             x
*            X
*            |
*           l|
*        j   |e    i
*       X    X    X
*
*               k
*              X
*
******************************************************/

void GetNewPoint(Point3D i, Point3D j, Point3D k, Coord l, Point3D x)
{
  Vector3D v1,v2,normal,v3;

  v1[X] = k[X] - i[X];
  v1[Y] = k[Y] - i[Y];
  v1[Z] = k[Z] - i[Z];

  v2[X] = i[X] - j[X];
  v2[Y] = i[Y] - j[Y];
  v2[Z] = i[Z] - j[Z];

/* interim result: normal (unnormalized) to plane i,j,k */
  GetVectorProduct(v1,v2,normal);
/* interim result: vector perpendicular to edge i,j */
  GetVectorProduct(normal,v2,v3);
/* normalize v3 */
  NormalizeVector(v3,v3);

/* end result: point with distance l to midpoint between i and j */
  x[X] = (0.5 * (i[X]+j[X])) + (l * v3[X]);
  x[Y] = (0.5 * (i[Y]+j[Y])) + (l * v3[Y]);
  x[Z] = (0.5 * (i[Z]+j[Z])) + (l * v3[Z]);

}/*GetNewPoint()*/


/******************************************************
*
* void MovePointOntoSurface(Point3D x)
*
* function moves point x onto isosurface (along gradient of x)
*
*          return: x
*
*                  x = x - f(x) * gradient(x)
*
******************************************************/

void MovePointOntoSurface(Point3D x)
{
  Coord    distance;
  Coord    epsilon = 0.01;
  Vector3D normal;

  distance = f(x[X],x[Y],x[Z]);

/* move point to isosurface until point is close enough */
  do
    {
      GetGradient(x,normal);
      x[X] = x[X] - (distance * normal[X]);
      x[Y] = x[Y] - (distance * normal[Y]);
      x[Z] = x[Z] - (distance * normal[Z]);
      distance = f(x[X],x[Y],x[Z]);
#ifdef DEBUG
  fprintf(stdout,"distance of point to isosurface: %2.3f\n",distance);
#endif
    } while( (distance > epsilon) || (distance < (-epsilon)) );

}/*MovePointOntoSurface()*/


/******************************************************
*
* Coord GetPointDistance(Point3D p1, Point3D p2)
*
* function returns distance between p1 and p3
*
******************************************************/

Coord GetPointDistance(Point3D p1, Point3D p2)
{
  Coord result;

  result = sqrt(((p1[X]-p2[X])*(p1[X]-p2[X])) +
                ((p1[Y]-p2[Y])*(p1[Y]-p2[Y])) +
                ((p1[Z]-p2[Z])*(p1[Z]-p2[Z])));

  return result;
}/*GetPointDistance()*/


/******************************************************
*
* int PointIsEqual(Point3D p1, Point3D p2)
*
* function checks p1 and p2 for equality
*
*          return: 0:     p1 != p2
*                  other: p1 == p2
*
******************************************************/

int PointIsEqual(Point3D p1, Point3D p2)
{
  return ((p1[X]==p2[X]) && (p1[Y]==p2[Y]) && (p1[Z]==p2[Z]));
}/*PointIsEqual()*/


/******************************************************
*
* void GetBoundingSphere(Point3D j, Point3D i, Point3D x,
*                        Point3D center, Coord *radius)
*
* function calculates bounding sphere of the triangle
*          described by j,i,x and returns the radius
*          and the center of the sphere
*
*          return: center = center of bounding sphere
*                  radius = radius of bounding sphere
*
*             x
*            X
*
*          center
*        j   X     i
*       X         X
*
*          center = intersection of line 1 and line 2
*
*          radius = distance between i and center
*
*
*          line 1 = described by point p1 (middle between i and j)
*                   and vector n1 (perpendicular to vector i-j and
*                   plane j,i,x)
*
*          line 1 = p1 + a1 * n1
*
*          line 2 = described by point p2 (middle between x and i)
*                   and vector n2 (perpendicular to vector x-i and
*                   plane j,i,x)
*
*          line 2 = p2 + a2 * n2
*
******************************************************/

void GetBoundingSphere(Point3D j, Point3D i, Point3D x,
                       Point3D center, Coord *radius)
{
  Vector3D v1;     /* vector (i - j) */
  Vector3D v2;     /* vector (x - i) */
/* p1 and p2 are stored as vectors to allow the usage of GetScalarProduct() */
  Vector3D vp1;    /* middle betwenn i and j */
  Vector3D vp2;    /* middle between x and i */
/* normals */
  Vector3D normal; /* vector perpendicular to plane j,i,x */
  Vector3D n1;     /* vector perpendicular to v1 and normal */
  Vector3D n2;     /* vector perpendicular to v2 and normal */

  Coord a2;        /* a2 from formula for line 2 */

/* calculate v1 */
  v1[X] = i[X] - j[X];
  v1[Y] = i[Y] - j[Y];
  v1[Z] = i[Z] - j[Z];
/* calculate v2 */
  v2[X] = x[X] - i[X];
  v2[Y] = x[Y] - i[Y];
  v2[Z] = x[Z] - i[Z];

/* calculate middle between i and j */
  vp1[X] = 0.5 * (i[X] + j[X]);
  vp1[Y] = 0.5 * (i[Y] + j[Y]);
  vp1[Z] = 0.5 * (i[Z] + j[Z]);
/* calculate middle between x and i */
  vp2[X] = 0.5 * (x[X] + i[X]);
  vp2[Y] = 0.5 * (x[Y] + i[Y]);
  vp2[Z] = 0.5 * (x[Z] + i[Z]);

/* calculate normals */
  GetVectorProduct(v1,v2,normal);
  GetVectorProduct(normal,v1,n1);
  GetVectorProduct(normal,v2,n2);
  NormalizeVector(n1,n1);
  NormalizeVector(n2,n2);

/* calculate a2 */
  a2 = ( GetScalarProduct(vp1,n2) * GetScalarProduct(n1,n1) +
         GetScalarProduct(n1,n2) * GetScalarProduct(vp2,n1) -
         GetScalarProduct(n1,n2) * GetScalarProduct(vp1,n1) -
         GetScalarProduct(n1,n1) * GetScalarProduct(vp2,n2) ) /
       ( GetScalarProduct(n1,n1) * GetScalarProduct(n2,n2) -
         GetScalarProduct(n1,n2) * GetScalarProduct(n2,n1) );

/* calculate center */
  center[X] = vp2[X] + (a2 * n2[X]);
  center[Y] = vp2[Y] + (a2 * n2[Y]);
  center[Z] = vp2[Z] + (a2 * n2[Z]);

/* calculate radius */
  *radius = (Coord) sqrt( (center[X] - i[X]) * (center[X] - i[X]) +
                          (center[Y] - i[Y]) * (center[Y] - i[Y]) +
                          (center[Z] - i[Z]) * (center[Z] - i[Z]) );

}/*GetBoundingSphere()*/


/******************************************************
*
* TriangleListElement * Create1stTriangle(Coord l);
*
* function creates first triangle needed for the 
*          marching triangle algorithm
*
*          return: pointer to triangle (list element)
*
*                  a point is chosen (somewhere) and 
*                  moved to the surface. this point is 
*                  moved a bit (in x) to get the second
*                  point. this point is moved a bit 
*                  (in -x and y). these point are the
*                  first triangle...
*
******************************************************/

TriangleListElement * Create1stTriangle(Coord l)
{
/* variables */
  Point3D p;
  Triangle t;
  TriangleListElement *start;

/* choose initial point somewhere */
  p[X] = 1.0; p[Y] = 2.0; p[Z] = 3.0;
  MovePointOntoSurface(p);
  t.i[X] = p[X]; t.i[Y] = p[Y]; t.i[Z] = p[Z];

/* move this point a bit to get second point */
  p[X] = p[X] + l;
  MovePointOntoSurface(p);
  t.j[X] = p[X]; t.j[Y] = p[Y]; t.j[Z] = p[Z];

/* move this point a bit to get third point */
  p[X] = p[X] - (0.5 * l); p[Y] = p[Y] + (0.7 * l);
  MovePointOntoSurface(p);
  t.k[X] = p[X]; t.k[Y] = p[Y]; t.k[Z] = p[Z];

/* create empty triangle stack */
  start = NULL;
/* init triangle stack by adding initila triangle to it */
  start = AddTriangleToListEnd(&t,start);

  return start;
}/*Create1stTriangle()*/


/******************************************************
*
* TriangleListElement * MarchingTriangle(Coord l, int maxTriangleCount, 
*                                        int fieldfunction);
*
* function returns pointer to triangle list describing
*          triangulated fieldfunction
*
*          l:                projection length of mt algorithm
*          maxTriangleCount: number of triangles to create (-1 = unlimited)
*          fieldfunction:    fieldfunction to be triangulated
*
******************************************************/

TriangleListElement * MarchingTriangle(Coord l, int maxTriangleCount, 
                                       int fieldfunction)
{
/* variables */

/* variables needed for heartbeat */
#ifndef DEBUG
  char progressChar[2] = "oO";
  int progressCounter;
  int progressCharID;
#endif
  
/* new points */
  Point3D xp;
  Point3D xpHalf;
  
/* new edges */
  Edge newLeftEdge;
  Edge newLeftEdge_2;
  Edge newRightEdge;
  Edge newRightEdge_2;

/* triangle / edge container */
  Triangle t;
  Edge edge;

/* cos(angle between edges) */
  Coord cosAngle;

/* length of newly created edge */
  Coord newEdgeLength;

/* radii of the bounding spheres */
  Coord radius_1;
  Coord radius_2;
/* center of the bounding spheres */
  Point3D center_1;
  Point3D center_2;

/* distances - needed for bounding sphere tests */
  Coord pointDistance_1;
  Coord pointDistance_2;

/* counter */
  int edgeCount;
  int triangleCount;
  int stepCount;
  int rejectCount;
  int removeCount;
  int longEdgeCount;
  int overLapCount;
  int edgeAge1Count;
  int edgeAge2Count;
  int edgeAge3Count;
  int oldEdgeCount;
  float midEdgeAge;
  int maxEdgeAge;

/* flags */
  int edgeOverLap;
  int overLapLeftEdge;
  int overLapRightEdge;
  int leftEdgeTooLong;
  int rightEdgeTooLong;
  int leftAngleIsSmall;
  int rightAngleIsSmall;
  int xpIsValid;
  int overLap;
  int boundingSphereType;

/* pointer */
  TriangleListElement *triangleStack    = NULL;

  EdgeListElement *edgeStack            = NULL;
  EdgeListElement *edgeIndex            = NULL;
  EdgeListElement *edgeToBeMoved        = NULL;

  EdgeListElement *activeEdge           = NULL;
  EdgeListElement *activeLeftEdge       = NULL;
  EdgeListElement *activeRightEdge      = NULL;

  EdgeListElement *newLeftEdgeAnchor    = NULL;
  EdgeListElement *newLeftEdgeAnchor_2  = NULL;
  EdgeListElement *newRightEdgeAnchor   = NULL;
  EdgeListElement *newRightEdgeAnchor_2 = NULL;

  PointListElement *pointList           = NULL;
  PointListElement *pointIndex          = NULL;


/***************** INIT MT ALGORITHM *****************/

/* some output... */
  fprintf(stdout,"\n");
  fprintf(stdout,"initialising mt...\n");

/* set fieldfunction */
  switch (fieldfunction)
    {
      case 1:
        f = f1;
        fprintf(stdout,"using fieldfunction 1 (sphere)\n");
        break;
      case 2:
        f = f2;
        fprintf(stdout,"using fieldfunction 2 (ellipsoid)\n");
        break;
      case 3:
        f = f3;
        fprintf(stdout,"using fieldfunction 3 (super-ellipsoid)\n");
        break;
      case 4:
        f = f4;
        fprintf(stdout,"using fieldfunction 4 (hot-dog)\n");
        break;
      case 5:
        f = f5;
        fprintf(stdout,"using fieldfunction 5 (buoy)\n");
        break;
      case 6:
        f = f6;
        fprintf(stdout,"using fieldfunction 6 (atom)\n");
        break;
      case 7:
        f = f7;
        fprintf(stdout,"using fieldfunction 7 (torus)\n");
        break;
      case 8:
        f = f8;
        fprintf(stdout,"using fieldfunction 8 (egg)\n");
        break;
      case 9:
        f = f9;
        fprintf(stdout,"using fieldfunction 9 (peanut)\n");
        break;
      default:
        f = f1;
        fprintf(stdout,"using default fieldfunction (sphere)\n");
        break;
    }/*switch*/


/* create initial triangle for mt */
  triangleStack = Create1stTriangle(l);

/* statistics */
  triangleCount = 1;
  stepCount     = 1;
  rejectCount   = 0;
  removeCount   = 0;
  longEdgeCount = 0;
  overLapCount  = 0;
  edgeAge1Count = 1;
  edgeAge2Count = 0;
  edgeAge3Count = 0;
  oldEdgeCount  = 0;
  midEdgeAge    = 0;
  maxEdgeAge    = 0;


/* create and store initial edges (all three views of the 1st triangle) */
/* edge3 */
  edge.i[X] = triangleStack->triangle.i[X];
  edge.i[Y] = triangleStack->triangle.i[Y];
  edge.i[Z] = triangleStack->triangle.i[Z];
  edge.j[X] = triangleStack->triangle.j[X];
  edge.j[Y] = triangleStack->triangle.j[Y];
  edge.j[Z] = triangleStack->triangle.j[Z];
  edge.k[X] = triangleStack->triangle.k[X];
  edge.k[Y] = triangleStack->triangle.k[Y];
  edge.k[Z] = triangleStack->triangle.k[Z];

  edgeStack = AddEdgeToListStart(&edge,edgeStack);
  activeEdge = edgeStack;

/* edge2 */
  edge.i[X] = triangleStack->triangle.j[X];                    
  edge.i[Y] = triangleStack->triangle.j[Y];
  edge.i[Z] = triangleStack->triangle.j[Z];
  edge.j[X] = triangleStack->triangle.k[X];
  edge.j[Y] = triangleStack->triangle.k[Y];
  edge.j[Z] = triangleStack->triangle.k[Z];
  edge.k[X] = triangleStack->triangle.i[X];
  edge.k[Y] = triangleStack->triangle.i[Y];
  edge.k[Z] = triangleStack->triangle.i[Z];

  edgeStack = AddEdgeToListStart(&edge,edgeStack);
  activeLeftEdge = edgeStack;

/* edge1 */
  edge.i[X] = triangleStack->triangle.k[X];
  edge.i[Y] = triangleStack->triangle.k[Y];
  edge.i[Z] = triangleStack->triangle.k[Z];
  edge.j[X] = triangleStack->triangle.i[X];
  edge.j[Y] = triangleStack->triangle.i[Y];
  edge.j[Z] = triangleStack->triangle.i[Z];
  edge.k[X] = triangleStack->triangle.j[X];
  edge.k[Y] = triangleStack->triangle.j[Y];
  edge.k[Z] = triangleStack->triangle.j[Z];

  edgeStack = AddEdgeToListStart(&edge,edgeStack);
  activeRightEdge = edgeStack;


/* init point list by adding vertices of 1st triangle */
  pointList = AddPointToListStart(&(edgeStack->edge.i),pointList);
  pointList = AddPointToListStart(&(edgeStack->edge.j),pointList);
  pointList = AddPointToListStart(&(edgeStack->edge.k),pointList);


/* create local dependencies */
  activeEdge->left       = activeLeftEdge;
  activeEdge->right      = activeRightEdge;
  activeLeftEdge->left   = activeRightEdge;
  activeLeftEdge->right  = activeEdge;
  activeRightEdge->left  = activeEdge;
  activeRightEdge->right = activeLeftEdge;

/* some output */
  fprintf(stdout,"\n");
  fprintf(stdout,"starting mt...\n");

/* prepare verbosity */
#ifndef DEBUG
  fprintf(stdout,"\n");
  fprintf(stdout,"working on triangle:          ");
  progressCharID = 0;
#endif


/************************* MT ************************/

/* mt will continue: as long there is an edge on the stack,
                     the number of triangles to be created has not 
                     been reached (unless the number is unlimited: -1),
                     there is an "unhandled" edge on the stack */
  while( (edgeStack != NULL) &&
         ((maxTriangleCount == -1) || (triangleCount < maxTriangleCount)) &&
         ((edgeStack->edge.id < triangleCount) || (edgeStack->edge.age < 3)) )
    {
/* statistics */
      stepCount++;

/* take first edge from stack */
      activeEdge      = edgeStack;
      activeLeftEdge  = activeEdge->left;
      activeRightEdge = activeEdge->right;

/* age: how many times has this edge been handled */
      activeEdge->edge.age = activeEdge->edge.age + 1;


/* init flags */
      edgeOverLap        = 0;
      overLapLeftEdge    = 0;
      overLapRightEdge   = 0;
      leftEdgeTooLong    = 0;
      rightEdgeTooLong   = 0;
      leftAngleIsSmall   = 0;
      rightAngleIsSmall  = 0;
      xpIsValid          = 0;
      overLap            = 0;
      boundingSphereType = STANDARD_BOUNDING_SPHERE;


/* verbosity */
#ifndef DEBUG
/* delete earlier printed output */
  fprintf(stdout,"%c%c%c%c%c%c%c%c%c",BS,BS,BS,BS,BS,BS,BS,BS,BS);
/* information for user */
  fprintf(stdout,"%6i  ",triangleCount+1);
/* heartbeat */
  progressCounter += 1;
  if(progressCounter > 50)
    {
      progressCounter = 0;
/* create toggle like this: 0 XOR 1 = 1 / 1 XOR 1 = 0 */
      progressCharID  = progressCharID ^ 1;
    }/*if*/
  fprintf(stdout,"%c",progressChar[progressCharID]);
  fflush(stdout);
#else
  fprintf(stdout,"\nworking on triangle %i, worked on this edge: %i times\n",
                    triangleCount+1,activeEdge->edge.age);
#endif


/******* debugging: check for data integrity  ********/

if(!(PointIsEqual(activeEdge->edge.j,activeLeftEdge->edge.i)))
  {
    fprintf(stderr,"\nerror in local dependencies: activeLeft wrong...\n");
    fprintf(stderr,"active:      i: %2.3f %2.3f %2.3f j: %2.3f %2.3f %2.3f\n",
                    activeEdge->edge.i[X],
                    activeEdge->edge.i[Y],
                    activeEdge->edge.i[Z],
                    activeEdge->edge.j[X],
                    activeEdge->edge.j[Y],
                    activeEdge->edge.j[Z]);
    fprintf(stderr,"activeLeft:  i: %2.3f %2.3f %2.3f j: %2.3f %2.3f %2.3f\n",
                    activeLeftEdge->edge.i[X],
                    activeLeftEdge->edge.i[Y],
                    activeLeftEdge->edge.i[Z],
                    activeLeftEdge->edge.j[X],
                    activeLeftEdge->edge.j[Y],
                    activeLeftEdge->edge.j[Z]);
    fprintf(stderr,"activeRight: i: %2.3f %2.3f %2.3f j: %2.3f %2.3f %2.3f\n",
                    activeRightEdge->edge.i[X],
                    activeRightEdge->edge.i[Y],
                    activeRightEdge->edge.i[Z],
                    activeRightEdge->edge.j[X],
                    activeRightEdge->edge.j[Y],
                    activeRightEdge->edge.j[Z]);
  }/*if*/

if(!(PointIsEqual(activeEdge->edge.i,activeRightEdge->edge.j)))
  {
    fprintf(stderr,"\nerror in local dependencies: activeRight wrong...\n");
    fprintf(stderr,"active:      i: %2.3f %2.3f %2.3f j: %2.3f %2.3f %2.3f\n",
                    activeEdge->edge.i[X],
                    activeEdge->edge.i[Y],
                    activeEdge->edge.i[Z],
                    activeEdge->edge.j[X],
                    activeEdge->edge.j[Y],
                    activeEdge->edge.j[Z]);
    fprintf(stderr,"activeLeft:  i: %2.3f %2.3f %2.3f j: %2.3f %2.3f %2.3f\n",
                    activeLeftEdge->edge.i[X],
                    activeLeftEdge->edge.i[Y],
                    activeLeftEdge->edge.i[Z],
                    activeLeftEdge->edge.j[X],
                    activeLeftEdge->edge.j[Y],
                    activeLeftEdge->edge.j[Z]);
    fprintf(stderr,"activeRight: i: %2.3f %2.3f %2.3f j: %2.3f %2.3f %2.3f\n",
                    activeRightEdge->edge.i[X],
                    activeRightEdge->edge.i[Y],
                    activeRightEdge->edge.i[Z],
                    activeRightEdge->edge.j[X],
                    activeRightEdge->edge.j[Y],
                    activeRightEdge->edge.j[Z]);
  }/*if*/


/**************** next / previous check **************/

/****************** check left angle *****************/

      cosAngle = GetEdgeAngle(activeEdge,activeLeftEdge);

/************ check for angle < MAX_ANGLE ************/

      if(cosAngle > COS_MAX_ANGLE)
        {
          edgeOverLap = 1;
          xp[X] = activeLeftEdge->edge.j[X];
          xp[Y] = activeLeftEdge->edge.j[Y];
          xp[Z] = activeLeftEdge->edge.j[Z];
#ifdef DEBUG
  fprintf(stdout,"left angle < MAX_ANGLE...\n");
#endif

/* check length of resulting new edge */
          newEdgeLength = GetPointDistance(xp,activeEdge->edge.i);
          if(newEdgeLength > (MAX_EDGE_LENGTH_FACTOR * l))
            {
              rightEdgeTooLong = 1;
#ifdef DEBUG
  fprintf(stdout,"new right edge too long: %2.3f times l\n",newEdgeLength/l);
#endif
              if(activeEdge->edge.age > 1)
                {
                  boundingSphereType = LONG_EDGE_BOUNDING_SPHERE;
#ifdef DEBUG
  fprintf(stdout,"old edge: using long edge bounding sphere...\n");
#endif
                }/*if*/
              else
                {
                  boundingSphereType = STANDARD_BOUNDING_SPHERE;
#ifdef DEBUG
  fprintf(stdout,"new edge: using standard bounding sphere...\n");
#endif
                }/*else*/
            }/*if(new edge too long)*/

/************ check for angle < SMALL_ANGLE **********/

          if(cosAngle > COS_SMALL_ANGLE)
            {
#ifdef DEBUG
  fprintf(stdout,"left angle < SMALL_ANGLE...\n");
#endif
              if(activeEdge->edge.age > 1)
                {
                  leftAngleIsSmall = 1;
                  boundingSphereType = SMALL_ANGLE_BOUNDING_SPHERE;
#ifdef DEBUG
  fprintf(stdout,"old edge: using small angle bounding sphere...\n");
#endif
                }/*if*/
              else
                {
                  boundingSphereType = STANDARD_BOUNDING_SPHERE;
#ifdef DEBUG
  fprintf(stdout,"new edge: using standard bounding sphere...\n");
#endif
                }/*else*/
            }/*if(small angle)*/
        }/*if(left angle < ...)*/

      else
/***************** check right angle *****************/
        {
          cosAngle = GetEdgeAngle(activeRightEdge,activeEdge);

/************ check for angle < MAX_ANGLE ************/

          if(cosAngle > COS_MAX_ANGLE)
            {
              edgeOverLap = 1;
              xp[X] = activeRightEdge->edge.i[X];
              xp[Y] = activeRightEdge->edge.i[Y];
              xp[Z] = activeRightEdge->edge.i[Z];
#ifdef DEBUG
  fprintf(stdout,"right angle < MAX_ANGLE...\n");
#endif

/* check length of resulting new edge */
              newEdgeLength = GetPointDistance(xp,activeEdge->edge.j);
              if(newEdgeLength > (MAX_EDGE_LENGTH_FACTOR * l))
                {
                  leftEdgeTooLong = 1;
#ifdef DEBUG
  fprintf(stdout,"new left edge too long: %2.3f times l\n",newEdgeLength/l);
#endif
                  if(activeEdge->edge.age > 1)
                    {
                      boundingSphereType = LONG_EDGE_BOUNDING_SPHERE;
#ifdef DEBUG
  fprintf(stdout,"old edge: using long edge bounding sphere...\n");
#endif
                    }/*if*/
                  else
                    {
                      boundingSphereType = STANDARD_BOUNDING_SPHERE;
#ifdef DEBUG
  fprintf(stdout,"new edge: using standard bounding sphere...\n");
#endif
                    }/*else*/
                }/*if(new edge too long)*/

/*********** check for angle < SMALL_ANGLE ***********/

              if(cosAngle > COS_SMALL_ANGLE)
                {
#ifdef DEBUG
  fprintf(stdout,"right angle < SMALL_ANGLE...\n");
#endif
                  if(activeEdge->edge.age > 1)
                    {
                      rightAngleIsSmall = 1;
                      boundingSphereType = SMALL_ANGLE_BOUNDING_SPHERE;
#ifdef DEBUG
  fprintf(stdout,"old edge, using small angle bounding sphere...\n");
#endif
                    }/*if*/
                  else
                    {
                      boundingSphereType = STANDARD_BOUNDING_SPHERE;
#ifdef DEBUG
  fprintf(stdout,"new edge: using standard bounding sphere...\n");
#endif
                    }/*else*/
                }/*if(small angle)*/
            }/*if(right angle < ...)*/
        }/*else*/


/***************** calculate new point ***************/

      if(!edgeOverLap)
        {
#ifdef DEBUG
  fprintf(stdout,"calculating new point...\n");
#endif
/* get new point */
          GetNewPoint(activeEdge->edge.i,
                      activeEdge->edge.j,
                      activeEdge->edge.k,
                      l,xp);
/* move the point onto the surface */
          MovePointOntoSurface(xp);
        }/*if*/

/* from now on xp is valid */
      xpIsValid = 1;


/*************** 3D Delaunay Constraint  *************/

/* more a check for overlap than a real Delaunay check */

/* calculate bounding sphere(s) of the to-be-triangle */
      switch(boundingSphereType)
        {
/* standard bounding sphere described by edge and xp */
          case STANDARD_BOUNDING_SPHERE :
            GetBoundingSphere(activeEdge->edge.i,
                              activeEdge->edge.j,
                              xp, center_1, &radius_1);
/* 2nd bounding sphere around xp to prevent creation of very small edges */
            if(!edgeOverLap)
              {
                center_2[X] = xp[X];
                center_2[Y] = xp[Y];
                center_2[Z] = xp[Z];
                radius_2    = (MIN_EDGE_LENGTH_FACTOR * l);
              }/*if*/
/* 2nd bounding sphere will be ignored if we are closing due to small angle */
            else
              {
                radius_2 = 0;
              }/*else*/
#ifdef DEBUG
  fprintf(stdout,"calculating bounding sphere (standard)...\n");
#endif
            break;/*STANDARD_BOUNDING_SPHERE*/
/* modified (smaller) bounding sphere for very long edges */
          case LONG_EDGE_BOUNDING_SPHERE :
            if(leftEdgeTooLong)
              {
                center_1[X] = 0.5 * (activeEdge->edge.j[X] +
                                     activeRightEdge->edge.i[X]);
                center_1[Y] = 0.5 * (activeEdge->edge.j[Y] +
                                     activeRightEdge->edge.i[Y]);
                center_1[Z] = 0.5 * (activeEdge->edge.j[Z] +
                                     activeRightEdge->edge.i[Z]);
                radius_1    = GetPointDistance(activeEdge->edge.j,center_1);
              }/*if*/
            if(rightEdgeTooLong)
              {
                center_1[X] = 0.5 * (activeEdge->edge.i[X] +
                                     activeLeftEdge->edge.j[X]);
                center_1[Y] = 0.5 * (activeEdge->edge.i[Y] +
                                     activeLeftEdge->edge.j[Y]);
                center_1[Z] = 0.5 * (activeEdge->edge.i[Z] +
                                     activeLeftEdge->edge.j[Z]);
                radius_1    = GetPointDistance(activeEdge->edge.i,center_1);
              }/*if*/
            radius_2 = 0;
#ifdef DEBUG
  fprintf(stdout,"calculating bounding sphere (long edge)...\n");
#endif
            break;/*LONG_EDGE_BOUNDING_SPHERE*/
/* modified (smaller) bounding sphere(s) to allow closing of small angles */
          case SMALL_ANGLE_BOUNDING_SPHERE :
            center_1[X] = 0.5 * (activeEdge->edge.i[X] +
                                 activeEdge->edge.j[X]);
            center_1[Y] = 0.5 * (activeEdge->edge.i[Y] +
                                 activeEdge->edge.j[Y]);
            center_1[Z] = 0.5 * (activeEdge->edge.i[Z] +
                                 activeEdge->edge.j[Z]);    
            radius_1    = GetPointDistance(activeEdge->edge.i,center_1);
/* 2nd bounding sphere */
            if(leftAngleIsSmall)
              {
                center_2[X] = 0.5 * (xp[X] + activeEdge->edge.j[X]);
                center_2[Y] = 0.5 * (xp[Y] + activeEdge->edge.j[Y]);
                center_2[Z] = 0.5 * (xp[Z] + activeEdge->edge.j[Z]);
              }/*if*/
            if(rightAngleIsSmall)
              {
                center_2[X] = 0.5 * (xp[X] + activeEdge->edge.i[X]);
                center_2[Y] = 0.5 * (xp[Y] + activeEdge->edge.i[Y]);
                center_2[Z] = 0.5 * (xp[Z] + activeEdge->edge.i[Z]);
              }/*if*/
            radius_2 = GetPointDistance(xp,center_2);
#ifdef DEBUG
  fprintf(stdout,"calculating bounding sphere (small angle)...\n");
#endif
            break;/*SMALL_ANGLE_BOUNDING_SPHERE*/
/* default -> error message */
          default:
            fprintf(stderr,"type of bounding sphere not set...\n");
        }/*switch(boundingSphereType)*/


/********* check for overlap (validity of xp) ********/

      pointIndex = pointList;

/* check whether there is an existing point in the bounding sphere(s) */
#ifdef DEBUG
  fprintf(stdout,"checking bounding sphere(s)...\n");
#endif
      while(pointIndex != NULL)
        {
          pointDistance_1 = GetPointDistance(center_1,pointIndex->point);
          pointDistance_2 = GetPointDistance(center_2,pointIndex->point);
/* one point is enough to invalidate xp */
          if( (pointDistance_1 < (0.999 * radius_1)) ||
              (pointDistance_2 < (0.999 * radius_2)) )
            {
/* unless the point is allowed to be in the bounding sphere depending on the 
   type of bounding sphere */
              switch(boundingSphereType)
                {
                  case STANDARD_BOUNDING_SPHERE :
/* if the edge has been handled several times, we try whether the overlap
   is a better choice than xp 
   an edgeoverlap has a higher priority! */
                    if((activeEdge->edge.age > 2) && !(edgeOverLap))
                      {
                        overLap = 1;
                        xp[X] = pointIndex->point[X];
                        xp[Y] = pointIndex->point[Y];
                        xp[Z] = pointIndex->point[Z];
#ifdef DEBUG
  fprintf(stdout,"overlap... validity will be checked later...\n");
#endif
                      }/*if*/
                    else
                      {
                        xpIsValid = 0;
#ifdef DEBUG
  fprintf(stdout,"xp invalid: edgeOverLap (standard)!\n");
#endif
                      }/*else*/
                    pointIndex = NULL;
                    break;/*STANDARD_BOUNDING_SPHERE*/
/* long resulting edge */
                  case LONG_EDGE_BOUNDING_SPHERE :
                    if( (leftEdgeTooLong &&
                         !PointIsEqual(pointIndex->point,
                                       activeEdge->edge.i) 
                        ) ||
                        (rightEdgeTooLong &&
                         !PointIsEqual(pointIndex->point,
                                       activeEdge->edge.j)
                        )
                      )
                      {
                        edgeOverLap = 1;
                        xpIsValid   = 0;
                        pointIndex  = NULL;
#ifdef DEBUG
  fprintf(stdout,"xp invalid: edgeOverLap (long edge)!\n");
#endif
                      }/*if*/
                    else
                      {
                        pointIndex = pointIndex->next;
                      }/*else*/
                    break;/*LONG_EDGE_BOUNDING_SPHERE*/
/* small angle */
                  case SMALL_ANGLE_BOUNDING_SPHERE :
                    if( ((leftAngleIsSmall) &&
                         !PointIsEqual(pointIndex->point,
                                       activeEdge->edge.i) &&
                         !PointIsEqual(pointIndex->point,
                                       activeEdge->edge.k) &&
                         !PointIsEqual(pointIndex->point,
                                       activeLeftEdge->edge.j) &&
                         !PointIsEqual(pointIndex->point,
                                       activeLeftEdge->edge.k)) ||
                        ((rightAngleIsSmall) &&
                         !PointIsEqual(pointIndex->point,
                                       activeEdge->edge.j) &&
                         !PointIsEqual(pointIndex->point,
                                       activeEdge->edge.k) &&
                         !PointIsEqual(pointIndex->point,
                                       activeRightEdge->edge.i) &&
                         !PointIsEqual(pointIndex->point,
                                       activeRightEdge->edge.k)) )
                      {
                        edgeOverLap = 1;                        
                        xpIsValid   = 0;
                        pointIndex  = NULL;
#ifdef DEBUG
  fprintf(stdout,"xp invalid: edgeOverLap (small angle)!\n");
#endif
                      }/*if*/
                    else
                      {
                        pointIndex = pointIndex->next;
                      }/*else*/
                    break;/*SMALL_ANGLE_BOUNDING_SPHERE*/
/* default -> error message */
                  default:
                    fprintf(stderr,"Type of Bounding Sphere not set...\n");
                }/*switch(boundingSphereType)*/
            }/*if*/
          else
            {
              pointIndex = pointIndex->next;
            }/*else*/
        }/*while*/


/********* check for validity of overlap *************/

/* if there was an overlap, a new check for overlap is needed */
      if(overLap)
        {
#ifdef DEBUG
  fprintf(stdout,"doing postponed validity-check...\n");
#endif
/* calculate new bounding sphere */
          GetBoundingSphere(activeEdge->edge.i,
                            activeEdge->edge.j,xp,
                            center_1,&radius_1);
          pointIndex = pointList;
/* check for overlap */
          while(pointIndex != NULL)
            {
              pointDistance_1 = GetPointDistance(center_1,pointIndex->point);
/* one point is enough to invalidate xp */
              if(pointDistance_1 < (0.999 * radius_1))
                {
                  xpIsValid = 0;
                  pointIndex = NULL;
#ifdef DEBUG
  fprintf(stdout,"xp invalid: overlap is invalid too...\n");
#endif
                }/*if*/
              else
                {
                  pointIndex = pointIndex->next;
                }/*else*/
            }/*while*/
        }/*if*/


/*** delete existing (if overlap) edges from stack ***/

      if((edgeOverLap) && (xpIsValid))
        {
/* left edge */
          if( (PointIsEqual(xp,activeLeftEdge->edge.j)) &&
              (PointIsEqual(activeEdge->edge.j,activeLeftEdge->edge.i)) &&
              !leftEdgeTooLong )
            {
              edgeIndex       = activeLeftEdge;
              overLapLeftEdge = 1;
              activeLeftEdge  = activeLeftEdge->left;
/* safety first: prevent access to NULL-pointer */
              if(edgeIndex->left != NULL)
                {
                  edgeIndex->left->right = NULL;          
                }/*if*/
              if(edgeIndex->right != NULL)
                {
                  edgeIndex->right->left = NULL;
                }/*if*/
/* statistics */
              removeCount++;
              edgeStack = RemoveEdgeListElement(edgeIndex,edgeStack);
              edgeIndex = NULL;
#ifdef DEBUG
  fprintf(stdout,"overLapLeftEdge...\n\n");
#endif
            }/*if*/
/* right edge */
          if( (PointIsEqual(xp,activeRightEdge->edge.i)) &&
              (PointIsEqual(activeEdge->edge.i,activeRightEdge->edge.j)) &&
              !rightEdgeTooLong )
            {
              edgeIndex        = activeRightEdge;
              overLapRightEdge = 1;
              activeRightEdge  = activeRightEdge->right;
/* safety first: prevent access to NULL-pointer */
              if(edgeIndex->left != NULL)
                {
                  edgeIndex->left->right = NULL;
                }/*if*/
              if(edgeIndex->right != NULL)
                {
                  edgeIndex->right->left = NULL;
                }/*if*/
/* statistics */
              removeCount++;
              edgeStack = RemoveEdgeListElement(edgeIndex,edgeStack);
              edgeIndex = NULL;
#ifdef DEBUG
  fprintf(stdout,"overLapRightEdge...\n\n");
#endif
            }/*if*/
        }/*if*/


/**************** create new triangle ****************/

/* create new triangle from two old and the new, now valid point */
      if(xpIsValid)
        {
/* statistics */
          if(activeEdge->edge.age > maxEdgeAge)
            {
              maxEdgeAge = activeEdge->edge.age;
            }/*if*/
          switch (activeEdge->edge.age)
            {
              case 1:
                edgeAge1Count++;
                break;
              case 2:
                edgeAge2Count++;
                break;
              case 3:
                edgeAge3Count++;
                break;
              default:
                oldEdgeCount++;
            }/*switch*/

/* if edges are normal sized: create new triangle */
          if(!(leftEdgeTooLong) && !(rightEdgeTooLong))
            {
/* old points */
              t.i[X] = activeEdge->edge.j[X];
              t.i[Y] = activeEdge->edge.j[Y];
              t.i[Z] = activeEdge->edge.j[Z];
              t.j[X] = activeEdge->edge.i[X];
              t.j[Y] = activeEdge->edge.i[Y];
              t.j[Z] = activeEdge->edge.i[Z];
/* new point */
              t.k[X] = xp[X];
              t.k[Y] = xp[Y];
              t.k[Z] = xp[Z];
/* store triangle on stack */
              triangleStack = AddTriangleToListStart(&t,triangleStack);
              triangleCount++;
/* add point to list */
              pointList = AddPointToListStart(&xp,pointList);
            }/*if*/
/* if one of the edges is too long: create two triangles instead of just one */
          else
            {
              if(leftEdgeTooLong)
                {
/* statistics */
                  longEdgeCount++;
/* bisect edge */
                  xpHalf[X] = 0.5 * (xp[X] + activeEdge->edge.j[X]);
                  xpHalf[Y] = 0.5 * (xp[Y] + activeEdge->edge.j[Y]);
                  xpHalf[Z] = 0.5 * (xp[Z] + activeEdge->edge.j[Z]);
                  MovePointOntoSurface(xpHalf);
                  t.i[X] = activeEdge->edge.i[X];
                  t.i[Y] = activeEdge->edge.i[Y];
                  t.i[Z] = activeEdge->edge.i[Z];
                  t.j[X] = xp[X];
                  t.j[Y] = xp[Y];
                  t.j[Z] = xp[Z];
                  t.k[X] = xpHalf[X];
                  t.k[Y] = xpHalf[Y];
                  t.k[Z] = xpHalf[Z];
/* store triangle on stack */
                  triangleStack = AddTriangleToListStart(&t,triangleStack);
                  triangleCount++;
                  t.i[X] = activeEdge->edge.j[X];
                  t.i[Y] = activeEdge->edge.j[Y];
                  t.i[Z] = activeEdge->edge.j[Z];
                  t.j[X] = activeEdge->edge.i[X];
                  t.j[Y] = activeEdge->edge.i[Y];
                  t.j[Z] = activeEdge->edge.i[Z];
/* store triangle on stack */
                  triangleStack = AddTriangleToListStart(&t,triangleStack);
                  triangleCount++;
/* add new points to list */
                  pointList = AddPointToListStart(&xp,pointList);
                  pointList = AddPointToListStart(&xpHalf,pointList);
                }/*if*/
              if(rightEdgeTooLong)
                {
/* statistics */
                  longEdgeCount++;
/* bisect edge */
                  xpHalf[X] = 0.5 * (xp[X] + activeEdge->edge.i[X]);
                  xpHalf[Y] = 0.5 * (xp[Y] + activeEdge->edge.i[Y]);
                  xpHalf[Z] = 0.5 * (xp[Z] + activeEdge->edge.i[Z]);
                  MovePointOntoSurface(xpHalf);
                  t.i[X] = xp[X];
                  t.i[Y] = xp[Y];
                  t.i[Z] = xp[Z];
                  t.j[X] = activeEdge->edge.j[X];
                  t.j[Y] = activeEdge->edge.j[Y];
                  t.j[Z] = activeEdge->edge.j[Z];
                  t.k[X] = xpHalf[X];
                  t.k[Y] = xpHalf[Y];
                  t.k[Z] = xpHalf[Z];
/* store triangle on stack */
                  triangleStack = AddTriangleToListStart(&t,triangleStack);
                  triangleCount++;
                  t.i[X] = activeEdge->edge.j[X];
                  t.i[Y] = activeEdge->edge.j[Y];
                  t.i[Z] = activeEdge->edge.j[Z];
                  t.j[X] = activeEdge->edge.i[X];
                  t.j[Y] = activeEdge->edge.i[Y];
                  t.j[Z] = activeEdge->edge.i[Z];
/* store triangle on stack */
                  triangleStack = AddTriangleToListStart(&t,triangleStack);
                  triangleCount++;
/* add new points to list */
                  pointList = AddPointToListStart(&xp,pointList);
                  pointList = AddPointToListStart(&xpHalf,pointList);
                }/*if*/
            }/*else*/
        }/*if(xpIsVaild)*/
/* move edge to back of stack if there is an overlap to get it solved later */
      else
        {
/* statistics */
          rejectCount++;
/* create ListElement from edge to be moved */
          edgeToBeMoved           = CreateEdgeListElement(&(edgeStack->edge));
          edgeToBeMoved->left     = activeLeftEdge;
          edgeToBeMoved->right    = activeRightEdge;
          edgeToBeMoved->edge.age = activeEdge->edge.age;
          edgeToBeMoved->edge.id  = triangleCount;
/* find end of list */
          edgeIndex = edgeStack;
          while(edgeIndex->next != NULL)
            {
              edgeIndex = edgeIndex->next;
            }/*while*/
          edgeIndex->next         = edgeToBeMoved;
          edgeToBeMoved->next     = NULL;
          edgeToBeMoved->previous = edgeIndex;
          edgeStack->left->right  = edgeToBeMoved;
          edgeStack->right->left  = edgeToBeMoved;
#ifdef DEBUG
  fprintf(stdout,"overlap: moving edge to end...\n");
#endif
        }/*else*/

/* remove handled edge from list */
      edgeStack = RemoveEdgeListElement(edgeStack,edgeStack);


/****************** create new edges *****************/

/* if xp is valid: create new edges and store them on stack */
      if(xpIsValid)
        {
/* create new edges */
/* right */
          if(!overLapRightEdge)
            {
              newRightEdge.i[X]  = triangleStack->triangle.j[X];
              newRightEdge.i[Y]  = triangleStack->triangle.j[Y];
              newRightEdge.i[Z]  = triangleStack->triangle.j[Z];
              newRightEdge.j[X]  = triangleStack->triangle.k[X];
              newRightEdge.j[Y]  = triangleStack->triangle.k[Y];
              newRightEdge.j[Z]  = triangleStack->triangle.k[Z];
              newRightEdge.k[X]  = triangleStack->triangle.i[X];
              newRightEdge.k[Y]  = triangleStack->triangle.i[Y];
              newRightEdge.k[Z]  = triangleStack->triangle.i[Z];
              edgeStack          = AddEdgeToListStart(&newRightEdge,edgeStack);
              newRightEdgeAnchor = edgeStack;
/* 2nd right if 1st was too long */
              if(rightEdgeTooLong)
                {
                  newRightEdge_2.i[X]  = triangleStack->next->triangle.k[X];
                  newRightEdge_2.i[Y]  = triangleStack->next->triangle.k[Y];
                  newRightEdge_2.i[Z]  = triangleStack->next->triangle.k[Z];
                  newRightEdge_2.j[X]  = triangleStack->next->triangle.i[X];
                  newRightEdge_2.j[Y]  = triangleStack->next->triangle.i[Y];
                  newRightEdge_2.j[Z]  = triangleStack->next->triangle.i[Z];
                  newRightEdge_2.k[X]  = triangleStack->next->triangle.j[X];
                  newRightEdge_2.k[Y]  = triangleStack->next->triangle.j[Y];
                  newRightEdge_2.k[Z]  = triangleStack->next->triangle.j[Z];
                  edgeStack            = AddEdgeToListStart(&newRightEdge_2,
                                                            edgeStack);
                  newRightEdgeAnchor_2 = edgeStack;
                }/*if*/
            }/*if*/
/* left */
          if(!overLapLeftEdge)
            {
              newLeftEdge.i[X]   = triangleStack->triangle.k[X];
              newLeftEdge.i[Y]   = triangleStack->triangle.k[Y];
              newLeftEdge.i[Z]   = triangleStack->triangle.k[Z];
              newLeftEdge.j[X]   = triangleStack->triangle.i[X];
              newLeftEdge.j[Y]   = triangleStack->triangle.i[Y];
              newLeftEdge.j[Z]   = triangleStack->triangle.i[Z];
              newLeftEdge.k[X]   = triangleStack->triangle.j[X];
              newLeftEdge.k[Y]   = triangleStack->triangle.j[Y];
              newLeftEdge.k[Z]   = triangleStack->triangle.j[Z];
              edgeStack          = AddEdgeToListStart(&newLeftEdge,edgeStack);
              newLeftEdgeAnchor  = edgeStack;
/* 2nd left if 1st was too long */
              if(leftEdgeTooLong)
                {
                  newLeftEdge_2.i[X]  = triangleStack->next->triangle.j[X];
                  newLeftEdge_2.i[Y]  = triangleStack->next->triangle.j[Y];
                  newLeftEdge_2.i[Z]  = triangleStack->next->triangle.j[Z];
                  newLeftEdge_2.j[X]  = triangleStack->next->triangle.k[X];
                  newLeftEdge_2.j[Y]  = triangleStack->next->triangle.k[Y];
                  newLeftEdge_2.j[Z]  = triangleStack->next->triangle.k[Z];
                  newLeftEdge_2.k[X]  = triangleStack->next->triangle.i[X];
                  newLeftEdge_2.k[Y]  = triangleStack->next->triangle.i[Y];
                  newLeftEdge_2.k[Z]  = triangleStack->next->triangle.i[Z];
                  edgeStack           = AddEdgeToListStart(&newLeftEdge_2,
                                                            edgeStack);
                  newLeftEdgeAnchor_2 = edgeStack;
                }/*if*/
            }/*if*/


/************ create local dependencies **************/

/* create new local dependencies */

/* no left overlap */
          if(!overLapLeftEdge)
            {
/* no left overlap / new left not too long */
              if(!leftEdgeTooLong)
                {
                  newLeftEdgeAnchor->left = activeLeftEdge;
                  activeLeftEdge->right   = newLeftEdgeAnchor;
/* no left overlap / new left not too long / new right not too long */
                  if(!overLapRightEdge)
                    {
                      newLeftEdgeAnchor->right  = newRightEdgeAnchor;
                      newRightEdgeAnchor->left  = newLeftEdgeAnchor;
                      newRightEdgeAnchor->right = activeRightEdge;
                      activeRightEdge->left     = newRightEdgeAnchor;
                    }/*if*/
/* no left overlap / new left not too long / new right too long */
                  else
                    {
                      newLeftEdgeAnchor->right = activeRightEdge;
                      activeRightEdge->left    = newLeftEdgeAnchor;
                    }/*else*/
                  }/*if*/
/* no left overlap / new left too long */
              else
                {
                  activeLeftEdge->right      = newLeftEdgeAnchor;
                  newLeftEdgeAnchor->right   = newLeftEdgeAnchor_2;
                  newLeftEdgeAnchor_2->right = activeRightEdge;
                  activeRightEdge->left      = newLeftEdgeAnchor_2;
                  newLeftEdgeAnchor_2->left  = newLeftEdgeAnchor;
                  newLeftEdgeAnchor->left    = activeLeftEdge;
                }/*else*/
            }/*if*/
/* left overlap */
          else
            {
/* left overlap / no right overlap */
              if(!overLapRightEdge)
                {
/* left overlap / no right overlap / new right not too long */
                  if(!rightEdgeTooLong)
                    {
                      newRightEdgeAnchor->left  = activeLeftEdge;
                      activeLeftEdge->right     = newRightEdgeAnchor;
                      newRightEdgeAnchor->right = activeRightEdge;
                      activeRightEdge->left     = newRightEdgeAnchor;
                    }/*if*/
/* left overlap / no right overlap / new right too long */
                  else
                    {
                      activeLeftEdge->right       = newRightEdgeAnchor_2;
                      newRightEdgeAnchor_2->right = newRightEdgeAnchor;
                      newRightEdgeAnchor->right   = activeRightEdge;
                      activeRightEdge->left       = newRightEdgeAnchor;
                      newRightEdgeAnchor->left    = newRightEdgeAnchor_2;
                      newRightEdgeAnchor_2->left  = activeLeftEdge;
                    }/*else*/
                }/*if*/
/* left overlap / right overlap */
              else
                {
                }/*else*/
            }/*else*/
        }/*if*/


/******* change local dependencies (if needed) *******/

    if(xpIsValid && overLap)
      {
/* count valid overlaps for statistics */
        overLapCount++;
        edgeIndex  = edgeStack;
/* find connected edges */
        while(edgeIndex != NULL)
          {
            if( (edgeIndex != newLeftEdgeAnchor) &&
                (PointIsEqual(xp,edgeIndex->edge.i)) )
              {
#ifdef DEBUG
  fprintf(stdout,"valid overlap... changing local dependencies!\n");
#endif
                edgeIndex->right->left = newLeftEdgeAnchor;
                newLeftEdgeAnchor->right = edgeIndex->right;
                edgeIndex->right = newRightEdgeAnchor;
                newRightEdgeAnchor->left = edgeIndex;
                edgeIndex = NULL;
              }/*if*/
            else
              {
                edgeIndex = edgeIndex->next;
              }/*else*/
          }/*while*/
      }/*if*/

/* next edge... */

    }/*while*/


/********************* END OF MT *********************/

/* some output */
  fprintf(stdout,"\n\n");
  fprintf(stdout,"finished with mt...\n");

/* delete point list thus freeing memory */
      while(pointList != NULL)
        {
          pointList = RemovePointListElement(pointList,pointList);
        }/*while*/


/******************* statistics **********************/

/* count unhandled edges and delete them */
  edgeCount = 0;
  while(edgeStack != NULL)
    {
      edgeCount++;
      edgeStack = RemoveEdgeListElement(edgeStack,edgeStack);
    }/*while*/
  fprintf(stdout,"\n");
  fprintf(stdout,"# edges left: %6i\n", edgeCount);

/* how many triangles are there */
  fprintf(stdout,"# triangles:  %6i\n", triangleCount);
  fprintf(stdout,"\n");

/* how many times were the edge handled in average */
  fprintf(stdout,"#edge age 1:  %6i\n", edgeAge1Count);
  fprintf(stdout,"#edge age 2:  %6i\n", edgeAge2Count);
  fprintf(stdout,"#edge age 3:  %6i\n", edgeAge3Count);
  fprintf(stdout,"#elder edges: %6i\n", oldEdgeCount);
  midEdgeAge = edgeAge1Count + (edgeAge2Count * 2) + (edgeAge3Count * 3);
  midEdgeAge = midEdgeAge / (edgeAge1Count + edgeAge2Count + edgeAge3Count);
  fprintf(stdout,"mid edge age: %6.2f\n", midEdgeAge);
  fprintf(stdout,"max edge age: %6i\n", maxEdgeAge);

  fprintf(stdout,"\n");

/* how many edges were handled */
  fprintf(stdout,"# steps:      %6i\n", stepCount);


/* how many edges were rejected */
  fprintf(stdout,"# rejects:    %6i\n", rejectCount);

/* how many edges were deleted from the stack due to closed small angles */
  fprintf(stdout,"# removes:    %6i\n", removeCount);

/* home many edges where too long and thus halfed */
  fprintf(stdout,"# long edges: %6i\n", longEdgeCount);

/* how many valid overlaps were there */
  fprintf(stdout,"# overlaps:   %6i\n", overLapCount);
  fprintf(stdout,"\n");

/******************** return *************************/

/* return pointer to list with all generated triangles to be drawn later */
  return triangleStack;
}/*MarchingTriangle()*/


/*****************************************************/

/* END_OF_FILE */
