/******************************************************
*
* copyright by H.G.Kleta, 2002, 2006
*
* if you wish to use these sources, please contact
*
* fhw at indigo2 dot de
*
******************************************************/

#include "scene.h"


/******************************************************
*
* variables
*
******************************************************/

/* light attributes */
#ifndef PRINTABLE
  float lightPos[]   = {-5, 0, 5, 0};
  float lightColor[] = {1, 1, 1, 1};
#else
  float lightPos[]   = {0, 0, 5, 0};
  float lightColor[] = {1, 1, 1, 1};
#endif

/* material attributes */
/* axis */
#ifndef NOAXIS
  float matAxisAmbient[RGBA]    = { 0.8, 0.8, 0.8, 1.0 };
  float matAxisDiffuse[RGBA]    = { 0.8, 0.8, 0.8, 1.0 };
  float matAxisSpecular[RGBA]   = { 0.0, 0.0, 0.0, 1.0 };
#endif
/* object */
  float matObjectAmbient[RGBA]  = { 1.0, 1.0, 1.0, 1.0 };
  float matObjectDiffuse[RGBA]  = { 1.0, 1.0, 1.0, 1.0 };
  float matObjectSpecular[RGBA] = { 0.0, 0.0, 0.0, 1.0 };
/* "global" */
float matShininess = 64;

/* textures */
GLfloat externalTextureData[TEXTURE_WIDTH][TEXTURE_HIGHT][RGBA];
GLfloat internalTextureData[TEXTURE_WIDTH][TEXTURE_HIGHT][RGBA];
GLfloat textureBorderColor[RGBA];

GLuint textureName[2];

/* ids for display list */
int CubeID;
int ObjectExtID;
int ObjectIntID;

/* used points */
Point3D p[] = {
              /* cube */
                { 1,  1,  1}, { -1,  1,  1}, { 1,  1, -1}, { -1,  1, -1},
                { 1, -1, -1}, { -1, -1, -1}, { 1, -1,  1}, { -1, -1,  1}
              };

/* rotation matrix */
GLfloat rotMatrix[16]  = { 1.0, 0.0, 0.0, 0.0,
                           0.0, 1.0, 0.0, 0.0,
                           0.0, 0.0, 1.0, 0.0,
                           0.0, 0.0, 0.0, 1.0 };


/******************************************************
*
* local function
*
* void InitTexture(void)
*
* function initialises texture
*
******************************************************/

void InitTexture(void)
{
#ifndef NOTEXTURE
/* texture coordinates */
  int i, j;

/* set texture border color */
#ifndef PRINTABLE
  textureBorderColor[R] = 1.0;
  textureBorderColor[G] = 1.0;
  textureBorderColor[B] = 1.0;
#else
  textureBorderColor[R] = 0.0;
  textureBorderColor[G] = 0.0;
  textureBorderColor[B] = 0.0;
#endif
  textureBorderColor[A] = 1.0;


/* uni color the texture */
  for(i = 0; i < TEXTURE_WIDTH; i++)
    {
      for(j = 0; j < TEXTURE_HIGHT; j++)
        {
#ifndef PRINTABLE
          externalTextureData[i][j][R] = 0.8;
          externalTextureData[i][j][G] = 0.7;
          externalTextureData[i][j][B] = 0.6;
#else
          externalTextureData[i][j][R] = 0.7;
          externalTextureData[i][j][G] = 0.7;
          externalTextureData[i][j][B] = 0.7;
#endif
          externalTextureData[i][j][A] = 1.0;

          internalTextureData[i][j][R] = 1.0;
          internalTextureData[i][j][G] = 1.0;
          internalTextureData[i][j][B] = 1.0;
          internalTextureData[i][j][A] = 0.3;
        }/*for*/
    }/*for*/


/* color for "wireframe" - diagonal */
  for(i = 0; i < TEXTURE_WIDTH; i++)
    {
      for(j = 0; j < TEXTURE_HIGHT; j++)
        {
          if((i == (j-2)) || (i == (j-1)) || (i == j) || (i == (j+1)))
            {
              externalTextureData[i][j][R] = textureBorderColor[R];
              externalTextureData[i][j][G] = textureBorderColor[G];
              externalTextureData[i][j][B] = textureBorderColor[B];

              internalTextureData[i][j][R] = textureBorderColor[R];
              internalTextureData[i][j][G] = textureBorderColor[G];
              internalTextureData[i][j][B] = textureBorderColor[B];
              internalTextureData[i][j][A] = 1.0;
            }
        }/*for*/
    }/*for*/

/* color for "wireframe" - horizontal */
  for(i = 0; i < TEXTURE_WIDTH; i++)
    {
      externalTextureData[i][TEXTURE_HIGHT-1][R] = textureBorderColor[R];
      externalTextureData[i][TEXTURE_HIGHT-1][G] = textureBorderColor[G];
      externalTextureData[i][TEXTURE_HIGHT-1][B] = textureBorderColor[B];

      externalTextureData[i][TEXTURE_HIGHT-2][R] = textureBorderColor[R];
      externalTextureData[i][TEXTURE_HIGHT-2][G] = textureBorderColor[G];
      externalTextureData[i][TEXTURE_HIGHT-2][B] = textureBorderColor[B];

      internalTextureData[i][TEXTURE_HIGHT-1][R] = textureBorderColor[R];
      internalTextureData[i][TEXTURE_HIGHT-1][G] = textureBorderColor[G];
      internalTextureData[i][TEXTURE_HIGHT-1][B] = textureBorderColor[B];
      internalTextureData[i][TEXTURE_HIGHT-1][A] = 1.0;

      internalTextureData[i][TEXTURE_HIGHT-2][R] = textureBorderColor[R];
      internalTextureData[i][TEXTURE_HIGHT-2][G] = textureBorderColor[G];
      internalTextureData[i][TEXTURE_HIGHT-2][B] = textureBorderColor[B];
      internalTextureData[i][TEXTURE_HIGHT-2][A] = 1.0;
    }/*for*/

/* color for "wireframe" - vertical */
  for(j = 0; j < TEXTURE_HIGHT; j++)
    {
      externalTextureData[0][j][R] = textureBorderColor[R];
      externalTextureData[0][j][G] = textureBorderColor[G];
      externalTextureData[0][j][B] = textureBorderColor[B];

      externalTextureData[1][j][R] = textureBorderColor[R];
      externalTextureData[1][j][G] = textureBorderColor[G];
      externalTextureData[1][j][B] = textureBorderColor[B];

      internalTextureData[0][j][R] = textureBorderColor[R];
      internalTextureData[0][j][G] = textureBorderColor[G];
      internalTextureData[0][j][B] = textureBorderColor[B];
      internalTextureData[0][j][A] = 1.0;

      internalTextureData[1][j][R] = textureBorderColor[R];
      internalTextureData[1][j][G] = textureBorderColor[G];
      internalTextureData[1][j][B] = textureBorderColor[B];
      internalTextureData[1][j][A] = 1.0;
    }/*for*/
#endif
}/*InitTexture()*/


/******************************************************
*
* local function
*
* void multMatrix(GLfloat *dest, GLfloat *src)
*
* function multiplies 4x4 matrices
*
******************************************************/

void multMatrix(GLfloat *dest, GLfloat *src)
{
/* copy of dest to avoid multiple usage of dest */
  GLfloat temp[16];
/* iteration counter */
  int i,j;

/* create copy of dest */
  for(i=0;i<16;++i)
    {
      temp[i]=dest[i];
    }/*for*/

/* do the actual work */
  for(i=0;i<4;++i)
    for(j=0;j<4;++j)
      {
        dest[i+4*j] = temp[0+4*j]*src[i] +
                      temp[1+4*j]*src[i+4] +
                      temp[2+4*j]*src[i+8] +
                      temp[3+4*j]*src[i+12];
      }/*for*/

}/*multMatrix()*/


/******************************************************
*
* int InitOpenGL(void)
*
* function returns 0 if the initialisation of the OpenGL
* environment fails, <> 0 if it succeeds
*
******************************************************/

int InitOpenGL(void)
{
  int success;

/* initialize and display window */
  glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
  glutInitWindowSize (600, 600);
  glutInitWindowPosition (0, 0);

  success = glutCreateWindow ("WWAIT: window without an interesting title");

  if (success)
    {
/* set color attributes */
#ifndef PRINTABLE
      glClearColor(0.0, 0.0, 0.0, 0.0);
      glColor4f(1.0, 1.0, 1.0, 1.0);
#else
      glClearColor(1.0, 1.0, 1.0, 0.0);
      glColor4f(0.0, 0.0, 0.0, 1.0);
#endif

/* enable back face culling */
      glEnable(GL_CULL_FACE);
      glCullFace(GL_BACK);

/* enable blending */
      glEnable(GL_BLEND);
      glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

/* enable automatic normalization of normal */
      glEnable(GL_NORMALIZE);

/* enable z-buffer */
      glEnable(GL_DEPTH_TEST);

/* "switch on" light */
      glEnable(GL_LIGHTING);
      glLightfv(GL_LIGHT0, GL_POSITION, lightPos);
      glLightfv(GL_LIGHT0, GL_DIFFUSE,  lightColor);
      glLightfv(GL_LIGHT0, GL_SPECULAR, lightColor);
      glEnable(GL_LIGHT0);

#ifndef NOTEXTURE
/* set texture attributes */
      InitTexture();
#endif

      glGenTextures(2, textureName);

/* external texture */
      glBindTexture(GL_TEXTURE_2D, textureName[0]);
      glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_WIDTH, TEXTURE_HIGHT,
                   0, GL_RGBA, GL_FLOAT, &externalTextureData[0][0][0]);
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, INTERPOL_METHOD);
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, INTERPOL_METHOD);
/* the texture border is needed, since GL_CLAMP_TO_BORDER is OpenGL 1.2
   and I want to be compatible with older installations */
      glTexParameterfv(GL_TEXTURE_2D, GL_TEXTURE_BORDER_COLOR,
                       textureBorderColor);


/* internal texture */
      glBindTexture(GL_TEXTURE_2D, textureName[1]);
      glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXTURE_WIDTH, TEXTURE_HIGHT,
                   0, GL_RGBA, GL_FLOAT, &internalTextureData[0][0][0]);
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, INTERPOL_METHOD);
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, INTERPOL_METHOD);
/* the texture border is needed, since GL_CLAMP_TO_BORDER is OpenGL 1.2
   and I want to be compatible with older installations */
      glTexParameterfv(GL_TEXTURE_2D, GL_TEXTURE_BORDER_COLOR,
                       textureBorderColor);


/* set viewattributes */
      glMatrixMode (GL_PROJECTION);
      gluPerspective(75, 1.0, 0.1, 100.0);
    }/*if*/

  return success;
}/*InitOpenGL()*/


/******************************************************
*
* void InitCube(void)
*
* function initialises a cube as OpenGL object
*          cube is used to construct axis
*
******************************************************/

void InitCube(void)
{
#ifndef NOAXIS
/* variable holding normal */
  Vector3D normal;

/* generate display-list */
  CubeID = glGenLists(1);

/* cube */
  glNewList(CubeID, GL_COMPILE);

/* top */
    glBegin(GL_POLYGON);
    GetNormal(p[0],p[2],p[3],normal);
    glNormal3fv((GLfloat *)normal);
    glVertex3fv((GLfloat *)p[0]); glVertex3fv((GLfloat *)p[2]);
    glVertex3fv((GLfloat *)p[3]); glVertex3fv((GLfloat *)p[1]);
    glEnd();

/* back */
    glBegin(GL_POLYGON);
    GetNormal(p[3],p[2],p[4],normal);
    glNormal3fv((GLfloat *)normal);
    glVertex3fv((GLfloat *)p[3]); glVertex3fv((GLfloat *)p[2]);
    glVertex3fv((GLfloat *)p[4]); glVertex3fv((GLfloat *)p[5]);
    glEnd();

/* bottom */
    glBegin(GL_POLYGON);
    GetNormal(p[5],p[4],p[6],normal);
    glNormal3fv((GLfloat *)normal);
    glVertex3fv((GLfloat *)p[5]); glVertex3fv((GLfloat *)p[4]);
    glVertex3fv((GLfloat *)p[6]); glVertex3fv((GLfloat *)p[7]);
    glEnd();

/* front */
    glBegin(GL_POLYGON);
    GetNormal(p[7],p[6],p[0],normal);
    glNormal3fv((GLfloat *)normal);
    glVertex3fv((GLfloat *)p[7]); glVertex3fv((GLfloat *)p[6]);
    glVertex3fv((GLfloat *)p[0]); glVertex3fv((GLfloat *)p[1]);
    glEnd();

/* right */
    glBegin(GL_POLYGON);
    GetNormal(p[0],p[6],p[4],normal);
    glNormal3fv((GLfloat *)normal);
    glVertex3fv((GLfloat *)p[0]); glVertex3fv((GLfloat *)p[6]);
    glVertex3fv((GLfloat *)p[4]); glVertex3fv((GLfloat *)p[2]);
    glEnd();

/* left */
    glBegin(GL_POLYGON);
    GetNormal(p[1],p[3],p[5],normal);
    glNormal3fv((GLfloat *)normal);
    glVertex3fv((GLfloat *)p[1]); glVertex3fv((GLfloat *)p[3]);
    glVertex3fv((GLfloat *)p[5]); glVertex3fv((GLfloat *)p[7]);
    glEnd();

  glEndList();
#endif
}/*InitCube()*/


/******************************************************
*
* void InitObject(TriangleListElement *start)
*
* function initialises object from triangles resulting 
*                      from marching triangle algorithm
*                      stored in list discribed by start
*
******************************************************/

void InitObject(TriangleListElement *start)
{
/* variable holding start of list*/
  TriangleListElement *triangleStack = start;

/* variable holding normal */
  Vector3D normal;

/* generate display-lists */
  ObjectExtID = glGenLists(1);
  ObjectIntID = glGenLists(1);


/* object - external view */
  glNewList(ObjectExtID, GL_COMPILE);

/* switch on texturing */
#ifndef NOTEXTURE
  glEnable(GL_TEXTURE_2D);
#endif

/* choose appropriate texture */
  glBindTexture(GL_TEXTURE_2D, textureName[0]);

/* set appropriate texturing mode */
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

/* walk through list and draw triangle */
  while(start != NULL)
    {
/* polygon model */
      glBegin(GL_TRIANGLES);
/* calculate normal for triangle */
        GetNormal((start->triangle.i),
                  (start->triangle.j),
                  (start->triangle.k),
                  normal);
/* apply normal */
        glNormal3fv((GLfloat *)normal);
/* create triangle from vertices */
        glTexCoord2f(1.0, 1.0); glVertex3fv((GLfloat *)(start->triangle.i));
        glTexCoord2f(0.0, 0.0); glVertex3fv((GLfloat *)(start->triangle.j));
        glTexCoord2f(1.0, 0.0); glVertex3fv((GLfloat *)(start->triangle.k));
      glEnd();
/* next triangle */
      start = start->next;
    }/*while*/

/* switch off texturing */
  glDisable(GL_TEXTURE_2D);

/* done */
  glEndList();


/* object - internal view */
  glNewList(ObjectIntID, GL_COMPILE);

/* switch on texturing */
#ifndef NOTEXTURE
  glEnable(GL_TEXTURE_2D);
#endif

/* choose appropriate texture */
  glBindTexture(GL_TEXTURE_2D, textureName[1]);

/* set appropriate texturing mode */
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

/* reset start */
  start = triangleStack;

/* walk through list and draw triangle */
  while(start != NULL)
    {
/* polygon model */
      glBegin(GL_TRIANGLES);
/* calculate normal for triangle */
        GetNormal((start->triangle.k),
                  (start->triangle.j),
                  (start->triangle.i),
                  normal);
/* apply normal */
        glNormal3fv((GLfloat *)normal);
/* create triangle from vertices */
        glTexCoord2f(1.0, 1.0); glVertex3fv((GLfloat *)(start->triangle.k));
        glTexCoord2f(0.0, 0.0); glVertex3fv((GLfloat *)(start->triangle.j));
        glTexCoord2f(1.0, 0.0); glVertex3fv((GLfloat *)(start->triangle.i));
      glEnd();                                              
/* next triangle */                                         
      start = start->next;
    }/*while*/

/* switch off texturing */
  glDisable(GL_TEXTURE_2D);

/* done */
  glEndList();


/* reset start */
  start = triangleStack;

/* walk through list and remove triangles, thus freeing memory */
  while(start != NULL)
    {
      start = RemoveTriangleListElement(start,start);
    }/*while*/

}/*InitObject()*/


/******************************************************
*
* void DrawAxis(void)
*
* function draws the coordinate axis
*
******************************************************/

void DrawAxis(void)
{
#ifndef NOAXIS
/* set material attributes */
  glMaterialfv(GL_FRONT, GL_AMBIENT,    matAxisAmbient);
  glMaterialfv(GL_FRONT, GL_DIFFUSE,    matAxisDiffuse);
  glMaterialfv(GL_FRONT, GL_SPECULAR,   matAxisSpecular);
  glMaterialfv(GL_FRONT, GL_SHININESS, &matShininess);

/* save matrix on stack as often as it is needed later on */
  glPushMatrix();
  glPushMatrix();
  glPushMatrix();
  glPushMatrix();
  glPushMatrix();

/* x-axis */
  glScalef( 5.5, 0.05, 0.05);
  glCallList(CubeID);

/* x */
  glPopMatrix();
  glTranslatef(5.5,0,0);
  glRotatef(90,0,1,0);
  glRotatef(45,0,0,1);
  glPushMatrix();
  glScalef( 0.5, 0.05, 0.05);
  glCallList(CubeID);
  glPopMatrix();
  glRotatef(90,0,0,1);
  glScalef( 0.5, 0.05, 0.05);
  glCallList(CubeID);

/* y-axis */
  glPopMatrix();
  glRotatef(90,0,0,1);
  glScalef( 5.5, 0.05, 0.05);
  glCallList(CubeID);

/* y */
  glPopMatrix();
  glTranslatef(0,5.5,0);
  glRotatef(90,1,0,0);
  glRotatef(-45,0,0,1);
  glPushMatrix();
  glScalef( 0.5, 0.05, 0.05);
  glCallList(CubeID);
  glPopMatrix();
  glRotatef(90,0,0,1);
  glTranslatef(-0.25,0,0);
  glScalef( 0.25, 0.05, 0.05);
  glCallList(CubeID);

/* z-axis */
  glPopMatrix();
  glRotatef(90,0,1,0);
  glScalef( 5.5, 0.05, 0.05);
  glCallList(CubeID);

/* z */
  glPopMatrix();
  glTranslatef(0,0,5.5);
  glRotatef(90,1,0,0);
  glPushMatrix();
  glPushMatrix();
  glRotatef(45,0,1,0);
  glScalef( 0.5, 0.05, 0.05);
  glCallList(CubeID);
  glPopMatrix();
  glTranslatef(0,0,-0.37);
  glScalef( 0.4, 0.05, 0.05);
  glCallList(CubeID);
  glPopMatrix();
  glTranslatef(0,0,0.37);
  glScalef( 0.4, 0.05, 0.05);
  glCallList(CubeID);
#endif
}/*DrawAxis()*/


/******************************************************
*
* void DrawObject(void)
*
* function draws all triangles from list initialised by
*          InitObject()
*
******************************************************/

void DrawObject(void)
{
/* make sure to draw the right size */
  glScalef( 1.0, 1.0, 1.0);

/* set material attributes */
  glMaterialfv(GL_FRONT, GL_AMBIENT,    matObjectAmbient);
  glMaterialfv(GL_FRONT, GL_DIFFUSE,    matObjectDiffuse);
  glMaterialfv(GL_FRONT, GL_SPECULAR,   matObjectSpecular);
  glMaterialfv(GL_FRONT, GL_SHININESS, &matShininess);

/* draw triangles - external view */
  glCallList(ObjectExtID);
/* draw triangles - internal view */
  glCallList(ObjectIntID);
}/*DrawObject()*/


/******************************************************
*
* void DrawWorld(void)
*
* function draws the graphical world
*
******************************************************/

void DrawWorld(void)
{
/* variables */

/* matrix for X rotation (init with id-matrix) */
  GLfloat rotXMatrix[16] = { 1.0, 0.0, 0.0, 0.0,
                             0.0, 1.0, 0.0, 0.0,
                             0.0, 0.0, 1.0, 0.0,
                             0.0, 0.0, 0.0, 1.0 };
/* matrix for Y rotation (init with id-matrix) */
  GLfloat rotYMatrix[16] = { 1.0, 0.0, 0.0, 0.0,
                             0.0, 1.0, 0.0, 0.0,
                             0.0, 0.0, 1.0, 0.0,
                             0.0, 0.0, 0.0, 1.0 };

/* initialize buffer */
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

/* init modeling matrix-stack */
  glMatrixMode(GL_MODELVIEW);

/* origin */
  glLoadIdentity();

/* set look-at 
   position can be changed with mouse */
  gluLookAt( -transX, -transY, transZ,
             -transX, -transY, transZ - 1.0,
             0, 1, 0);

/* set rotation matrices */
  rotXMatrix[5]  =  cos(-rotX * PI / 180);
  rotXMatrix[6]  = -sin(-rotX * PI / 180);
  rotXMatrix[9]  =  sin(-rotX * PI / 180);
  rotXMatrix[10] =  cos(-rotX * PI / 180);

  rotYMatrix[0]  =  cos(-rotY * PI / 180);
  rotYMatrix[2]  =  sin(-rotY * PI / 180);
  rotYMatrix[8]  = -sin(-rotY * PI / 180);
  rotYMatrix[10] =  cos(-rotY * PI / 180);


  if(!resetScene)
    {
/* calculate actual rotation matrix */
      multMatrix((GLfloat*)rotMatrix,(GLfloat*)rotXMatrix);
      multMatrix((GLfloat*)rotMatrix,(GLfloat*)rotYMatrix);
    }/*if*/
  else
/* reset rotation matrix to id-matrix if wanted thus resetting scene */
    {
      rotMatrix[0]  = 1.0;
      rotMatrix[1]  = 0.0;
      rotMatrix[2]  = 0.0;
      rotMatrix[3]  = 0.0;
      rotMatrix[4]  = 0.0;
      rotMatrix[5]  = 1.0;
      rotMatrix[6]  = 0.0;
      rotMatrix[7]  = 0.0;
      rotMatrix[8]  = 0.0;
      rotMatrix[9]  = 0.0;
      rotMatrix[10] = 1.0;
      rotMatrix[11] = 0.0;
      rotMatrix[12] = 0.0;
      rotMatrix[13] = 0.0;
      rotMatrix[14] = 0.0;
      rotMatrix[15] = 1.0;
      resetScene = 0.0;
    }/*else*/

/* apply rotation matrix */
  glMultMatrixf((GLfloat*)rotMatrix);

/* rotation is done: reset values */
  rotX = 0.0;
  rotY = 0.0;


/* save matrix on stack */
  glPushMatrix();

/* draw the x,y and z axis */
  DrawAxis();

/* restore the matrix from the stack */
  glPopMatrix();

/* draw result of the marching triangle algorithm */
  DrawObject();

/* display objects */
  glFlush();

/* swap front- / back-buffer */
  glutSwapBuffers();
}/*DrawWorld()*/


/*****************************************************/

/* END OF FILE */
